from pettingzoo.utils.depreciated_module import DepreciatedModule

simple_adversary_v0 = DepreciatedModule("simple_adversary", "v0", "v2")
simple_adversary_v1 = DepreciatedModule("simple_adversary", "v1", "v2")
simple_crypto_v0 = DepreciatedModule("simple_crypto", "v0", "v2")
simple_crypto_v1 = DepreciatedModule("simple_crypto", "v1", "v2")
simple_push_v0 = DepreciatedModule("simple_push", "v0", "v2")
simple_push_v1 = DepreciatedModule("simple_push", "v1", "v2")
simple_reference_v0 = DepreciatedModule("simple_reference", "v0", "v2")
simple_reference_v1 = DepreciatedModule("simple_reference", "v0", "v2")

simple_speaker_listener_v0 = DepreciatedModule("simple_speaker_listener", "v0", "v3")
simple_speaker_listener_v1 = DepreciatedModule("simple_speaker_listener", "v1", "v3")
simple_speaker_listener_v2 = DepreciatedModule("simple_speaker_listener", "v2", "v3")

simple_spread_v0 = DepreciatedModule("simple_spread", "v0", "v2")
simple_spread_v1 = DepreciatedModule("simple_spread", "v1", "v2")
simple_tag_v0 = DepreciatedModule("simple_tag", "v0", "v2")
simple_tag_v1 = DepreciatedModule("simple_tag", "v1", "v2")
simple_v0 = DepreciatedModule("simple", "v0", "v2")
simple_v1 = DepreciatedModule("simple", "v1", "v2")
simple_world_comm_v0 = DepreciatedModule("simple_world_comm", "v0", "v2")
simple_world_comm_v1 = DepreciatedModule("simple_world_comm", "v1", "v2")
