from pettingzoo.utils.depreciated_module import DepreciatedModule

basketball_pong_v0 = DepreciatedModule("basketball_pong", "v0", "v1")
boxing_v0 = DepreciatedModule("boxing", "v0", "v1")
combat_plane_v0 = DepreciatedModule("combat_plane", "v0", "v1")
combat_tank_v0 = DepreciatedModule("combat_tank", "v0", "v1")
double_dunk_v0 = DepreciatedModule("double_dunk", "v0", "v2")
double_dunk_v1 = DepreciatedModule("double_dunk", "v1", "v2")
entombed_competitive_v0 = DepreciatedModule("entombed_competitive", "v0", "v2")
entombed_competitive_v1 = DepreciatedModule("entombed_competitive", "v1", "v2")
entombed_cooperative_v0 = DepreciatedModule("entombed_cooperative", "v0", "v2")
entombed_cooperative_v1 = DepreciatedModule("entombed_cooperative", "v1", "v2")
flag_capture_v0 = DepreciatedModule("flag_capture", "v0", "v1")
foozpong_v0 = DepreciatedModule("foozpong", "v0", "v1")
ice_hockey_v0 = DepreciatedModule("ice_hockey", "v0", "v1")
joust_v0 = DepreciatedModule("joust", "v0", "v2")
joust_v1 = DepreciatedModule("joust", "v1", "v2")
mario_bros_v0 = DepreciatedModule("mario_bros", "v0", "v2")
mario_bros_v1 = DepreciatedModule("mario_bros", "v1", "v2")
maze_craze_v0 = DepreciatedModule("maze_craze", "v0", "v2")
maze_craze_v1 = DepreciatedModule("maze_craze", "v1", "v2")
othello_v0 = DepreciatedModule("othello", "v0", "v2")
othello_v1 = DepreciatedModule("othello", "v1", "v2")
pong_v0 = DepreciatedModule("pong", "v0", "v1")
quadrapong_v0 = DepreciatedModule("quadrapong", "v0", "v2")
quadrapong_v1 = DepreciatedModule("quadrapong", "v1", "v2")
space_invaders_v0 = DepreciatedModule("space_invaders", "v0", "v1")
space_war_v0 = DepreciatedModule("space_war", "v0", "v1")
surround_v0 = DepreciatedModule("surround", "v0", "v1")
tennis_v0 = DepreciatedModule("tennis", "v0", "v2")
tennis_v1 = DepreciatedModule("tennis", "v1", "v2")
video_checkers_v0 = DepreciatedModule("video_checkers", "v0", "v2")
video_checkers_v1 = DepreciatedModule("video_checkers", "v1", "v2")
volleyball_pong_v0 = DepreciatedModule("volleyball_pong", "v0", "v1")
warlords_v0 = DepreciatedModule("warlords", "v0", "v2")
warlords_v1 = DepreciatedModule("warlords", "v1", "v2")
wizard_of_wor_v0 = DepreciatedModule("wizard_of_wor", "v0", "v2")
wizard_of_wor_v1 = DepreciatedModule("wizard_of_wor", "v1", "v2")
