from pettingzoo.utils.depreciated_module import DepreciatedModule

prison_v0 = DepreciatedModule("prison", "v0", "v2")
prison_v1 = DepreciatedModule("prison", "v1", "v2")
prospector_v0 = DepreciatedModule("prospector", "v0", "v3")
prospector_v1 = DepreciatedModule("prospector", "v1", "v3")
prospector_v2 = DepreciatedModule("prospector", "v2", "v3")
pistonball_v0 = DepreciatedModule("pistonball", "v0", "v2")
pistonball_v1 = DepreciatedModule("pistonball", "v1", "v2")
cooperative_pong_v0 = DepreciatedModule("cooperative_pong", "v0", "v2")
cooperative_pong_v1 = DepreciatedModule("cooperative_pong", "v1", "v2")
knights_archers_zombies_v0 = DepreciatedModule("knights_archers_zombies", "v0", "v4")
knights_archers_zombies_v1 = DepreciatedModule("knights_archers_zombies", "v1", "v4")
knights_archers_zombies_v2 = DepreciatedModule("knights_archers_zombies", "v2", "v4")
knights_archers_zombies_v3 = DepreciatedModule("knights_archers_zombies", "v3", "v4")
