from pettingzoo.utils.depreciated_module import DepreciatedModule

adversarial_pursuit_v0 = DepreciatedModule("adversarial_pursuit", "v0", "v2")
adversarial_pursuit_v1 = DepreciatedModule("adversarial_pursuit", "v1", "v2")
battle_v0 = DepreciatedModule("battle", "v0", "v2")
battle_v1 = DepreciatedModule("battle", "v1", "v2")
battlefield_v0 = DepreciatedModule("battlefield", "v0", "v2")
battlefield_v1 = DepreciatedModule("battlefield", "v1", "v2")
combined_arms_v0 = DepreciatedModule("combined_arms", "v0", "v2")
combined_arms_v1 = DepreciatedModule("combined_arms", "v1", "v2")
gather_v0 = DepreciatedModule("gather", "v0", "v2")
gather_v1 = DepreciatedModule("gather", "v1", "v2")
tiger_deer_v0 = DepreciatedModule("tiger_deer", "v0", "v2")
tiger_deer_v1 = DepreciatedModule("tiger_deer", "v1", "v2")
