from pettingzoo.utils.depreciated_module import DepreciatedModule

multiwalker_v0 = DepreciatedModule("multiwalker", "v0", "v4")
multiwalker_v1 = DepreciatedModule("multiwalker", "v1", "v4")
multiwalker_v2 = DepreciatedModule("multiwalker", "v2", "v4")
multiwalker_v3 = DepreciatedModule("multiwalker", "v3", "v4")
pursuit_v0 = DepreciatedModule("pursuit", "v0", "v2")
pursuit_v1 = DepreciatedModule("pursuit", "v1", "v2")
waterworld_v0 = DepreciatedModule("waterworld", "v0", "v2")
waterworld_v1 = DepreciatedModule("waterworld", "v1", "v2")
